/* 1508, Tue 6 Jun 95

   FLOWDATA.H:  Global declarations for fd_filter and fd_extract

   Copyright (C) 1994-1995 by Nevil Brownlee,
   Computer Centre,  The University of Auckland */

#ifndef EXTERN
#define EXTERN  extern
#define DECLARE  0
#define INIT(v)
#else
#define EXTERN
#define DECLARE  1
#define INIT(v)  = v
#endif

extern int errno;

EXTERN int snmp_dump_packet INIT(0);

EXTERN int verbose, testing, listrules, listformat, quiet;

#define MXTAGS  200
#define MXCOLS   20

#define MAC_ADDR_LEN      6
#define PEER_ADDR_LEN    20
#define DETAIL_ADDR_LEN   2

#define RULE_ADDR_LEN    20

#define string_value(a)  (attribs[a].len > 4)

#define STR_LEN   64

struct flow_info {  /* Accounting flow as viewed by programs */
   unsigned int FlowIndex;
   unsigned int LowInterface, HighInterface;
   unsigned int LowAdjType, HighAdjType;
   unsigned char LowAdjAddress[STR_LEN+1], LowAdjMask[STR_LEN+1],
      HighAdjAddress[STR_LEN+1], HighAdjMask[STR_LEN+1];
   unsigned int
      LowPeerType, HighPeerType;
   unsigned char LowPeerAddress[STR_LEN+1], LowPeerMask[STR_LEN+1],
      HighPeerAddress[STR_LEN+1], HighPeerMask[STR_LEN+1];
   unsigned int
      LowDetailType, HighDetailType;
   unsigned int   LowDetailAddress, LowDetailMask,
      HighDetailAddress, HighDetailMask;
   unsigned int FlowStatus, FlowRuleSet;
   unsigned int
      FwdPackets,FwdBytes,  BackPackets,BackBytes,
      FirstTime,LastTime;
   unsigned int LowClass, HighClass, FlowClass;
   unsigned int LowKind, HighKind, FlowKind;

   unsigned int  /* New attributes (created by fd_filter) */
      upbci,   /* To byte rate (bytes since last sample) */
      dnbci,   /* From byte rate */
      uppci,   /* To packet rate */
      dnpci,   /* From packet rate */
      TagNbr;  /* Tag nbr */
   };
EXTERN struct flow_info curr_flow;


#define AT_IGNORE     0  /* Addr_type values */
#define AT_ADJACENT   1
#define AT_IP         2
#define AT_CLNS       3
#define AT_IDPR       4
#define AT_DECNET     5
#define AT_NOVELL     6
#define AT_ETHERTALK  7
#define AT_DETAIL     8
#define AT_IP_FRAG   11
#define AT_OTHER     12

#define PT_ICMP       1  /* IP protocol type values */
#define PT_TCP        6
#define PT_UDP       17
#define PT_OSPF      89

#define WNP_FTPDATA  20  /* Well-known tcp/udp port numbers */
#define WNP_FTP      21
#define WNP_TELNET   23
#define WNP_SMTP     25
#define WNP_DOMAIN   53
#define WNP_GOPHER   70
#define WNP_WWW      80
#define WNP_NNTP    119
#define WNP_NTP     123
#define WNP_SNMP    161

#define RF_SET       -2  /* -1 is EOF !!! */
#define RF_RULES     -3
#define RF_ACTIONS   -4
#define RF_FORMAT    -5
#define RF_STATS     -6
#define RF_TAG       -7
#define RF_COLUMN    -8
#define RF_TIME      -9

#define EX_SCALE    -51  /* Keywords for fd_extract */
#define EX_KILO     -52
#define EX_MEGA     -53
#define EX_ELAPSED  -54
#define EX_CLOCK    -55
#define EX_SECS     -56
#define EX_MINS     -57
#define EX_HOURS    -58
#define EX_DAYS     -59

#define FTFLOWINDEX		1  /* Flow table attribute values */
#define FTFLOWSTATUS		2

#define FTLOWINTERFACE          3
#define FTLOWADJACENTTYPE       4
#define FTLOWADJACENTADDRESS    5
#define FTLOWADJACENTMASK       6
#define FTLOWPEERTYPE		7
#define FTLOWPEERADDRESS	8
#define FTLOWPEERMASK	        9
#define FTLOWDETAILTYPE        10
#define FTLOWDETAILADDRESS     11
#define FTLOWDETAILMASK        12

#define FTHIINTERFACE          13
#define FTHIADJACENTTYPE       14
#define FTHIADJACENTADDRESS    15
#define FTHIADJACENTMASK       16
#define FTHIPEERTYPE	       17
#define FTHIPEERADDRESS	       18
#define FTHIPEERMASK	       19
#define FTHIDETAILTYPE         20
#define FTHIDETAILADDRESS      21
#define FTHIDETAILMASK         22

#define FTPDUSCALE             23
#define FTOCTETSCALE           24
#define FTRULESET              25

#define FTUPOCTETS	       26
#define FTUPPDUS	       27
#define FTDOWNOCTETS	       28
#define FTDOWNPDUS	       29
#define FTFIRSTTIME	       30
#define FTLASTTIME	       31

#define FTLOWSUBSCRIBERID      32
#define FTHISUBSCRIBERID       33
#define FTSESSIONID            34

#define FTSOURCECLASS          35
#define FTDESTCLASS            36
#define FTFLOWCLASS            37
#define FTSOURCEKIND           38
#define FTDESTKIND             39
#define FTFLOWKIND             40

#define FTUPOCTETRATE          41  /* Attributes added by fd_filter */
#define FTUPPDURATE	       42
#define FTDOWNOCTETRATE	       43
#define FTDOWNPDURATE	       44
#define FTTAGNBR               45

#define NATTRIBS         FTTAGNBR

typedef union {
   unsigned int *intval;
   unsigned char *charval;
   } val;
struct attrib_info {
   char *name;
   unsigned char index;
   unsigned char len;
   val value;
   };

#define SZ_ATTRIBS  (1+NATTRIBS+6)
EXTERN struct attrib_info attribs[SZ_ATTRIBS]
#if DECLARE
   = {
   "null",                   0,                     0,  NULL,
   "flowindex",              FTFLOWINDEX,           2,  NULL,
   "flowstatus",             FTFLOWSTATUS,          1,  NULL,
   "sourceinterface",        FTLOWINTERFACE,        1,  NULL,
   "sourceadjacenttype",     FTLOWADJACENTTYPE,     1,  NULL,
   "sourceadjacentaddress",  FTLOWADJACENTADDRESS,  MAC_ADDR_LEN,  NULL,
   "sourceadjacentmask",     FTLOWADJACENTMASK,     MAC_ADDR_LEN,  NULL,
   "sourcepeertype",         FTLOWPEERTYPE,         1,  NULL,
   "sourcepeeraddress",      FTLOWPEERADDRESS,      PEER_ADDR_LEN,  NULL,
   "sourcepeermask",         FTLOWPEERMASK,         PEER_ADDR_LEN,  NULL,
   "sourcetranstype",        FTLOWDETAILTYPE,       1,  NULL,
   "sourcetransaddress",     FTLOWDETAILADDRESS,    DETAIL_ADDR_LEN,  NULL,
   "sourcetransmask",        FTLOWDETAILMASK,       DETAIL_ADDR_LEN,  NULL,

   "destinterface",          FTHIINTERFACE,         1,  NULL,
   "destadjacenttype",       FTHIADJACENTTYPE,      1,  NULL,
   "destadjacentaddress",    FTHIADJACENTADDRESS,   MAC_ADDR_LEN,  NULL,
   "destadjacentmask",       FTHIADJACENTMASK,      MAC_ADDR_LEN,  NULL,
   "destpeertype",           FTHIPEERTYPE,          1,  NULL,
   "destpeeraddress",        FTHIPEERADDRESS,       PEER_ADDR_LEN,  NULL,
   "destpeermask",           FTHIPEERMASK,          PEER_ADDR_LEN,  NULL,
   "desttranstype",          FTHIDETAILTYPE,        1,  NULL,
   "desttransaddress",       FTHIDETAILADDRESS,     DETAIL_ADDR_LEN,  NULL,
   "desttransmask",          FTHIDETAILMASK,        DETAIL_ADDR_LEN,  NULL,

   "pduscale",               FTPDUSCALE,            0,  NULL,
   "octetscale",             FTOCTETSCALE,          0,  NULL,
   "flowruleset",            FTRULESET,             1,  NULL,
   "tooctets",               FTUPOCTETS,            4,  NULL,
   "topdus",                 FTUPPDUS,              4,  NULL,
   "fromoctets",             FTDOWNOCTETS,          4,  NULL,
   "frompdus",               FTDOWNPDUS,            4,  NULL,
   "firsttime",              FTFIRSTTIME,           4,  NULL,
   "lasttime",               FTLASTTIME,            4,  NULL,

   "sourcesubscriberid",     FTLOWSUBSCRIBERID,     0,  NULL,
   "destsubscriberid",       FTHISUBSCRIBERID,      0,  NULL,
   "sessionid",              FTSESSIONID,           0,  NULL,

   "sourceclass",            FTSOURCECLASS,         1,  NULL,
   "destclass",              FTDESTCLASS,           1,  NULL,
   "flowclass",              FTFLOWCLASS,           1,  NULL,
   "sourcekind",             FTSOURCEKIND,          1,  NULL,
   "destkind",               FTDESTKIND,            1,  NULL,
   "flowkind",               FTFLOWKIND,            1,  NULL,

   "tooctetrate",            FTUPOCTETRATE,         4,  NULL,
   "topdurate",              FTUPPDURATE,           4,  NULL,
   "fromoctetrate",          FTDOWNOCTETRATE,       4,  NULL,
   "frompdurate",            FTDOWNPDURATE,         4,  NULL,
   "tagnbr",		     FTTAGNBR,		    4,  NULL,

   "sourcedetailtype",       FTLOWDETAILTYPE,       1,	NULL,
   "sourcedetailaddress",    FTLOWDETAILADDRESS,    DETAIL_ADDR_LEN,  NULL,
   "sourcedetailmask",       FTLOWDETAILMASK,       DETAIL_ADDR_LEN,  NULL,
   "destdetailtype",         FTHIDETAILTYPE,        1,	NULL,
   "destdetailaddress",      FTHIDETAILADDRESS,     DETAIL_ADDR_LEN,  NULL,
   "destdetailmask",         FTHIDETAILMASK,        DETAIL_ADDR_LEN,  NULL
   }
#endif
   ;

struct flow_stats {

   unsigned char format[1+NATTRIBS];
   char *separator[1+NATTRIBS];
   unsigned char required[1+NATTRIBS];

   FILE *flows;
   };



EXTERN FILE *rfp;  /* Scanner globals */
EXTERN char inbuf[500],*ibp, ebuf[80];  /* Input, error message buffers */
EXTERN int lic, ic,  /* Last input char, current input char */
   iblisted, rule_line, rferrors;


/* Forward procedure declarations .. */

/* Functions declared in fd_parse.c */

char *gnbr(unsigned int *n, char *s);
char *gcstring(char *s, int *len);

int parse_open(char *fn);
void p_error(char *msg);
int nextchar(void);
int nexttoken(void);
void getfdstring(unsigned char *arg);
int wordis(char *p, char *w);
int getword(void);
int getnbr(void);

void scan_format(struct flow_stats *fs, int single_record, int list);

val attrib_ptr(struct flow_info *fip, int a);
char *get_value(val v, int a, char *pb);
void put_value(FILE *out, int a, val v);
